from typing import List, Set, Dict, Union
import numpy as np
import pandas as pd
from numpy.core.multiarray import ndarray
from quicksectx import IntervalTree
from spacy.tokens.doc import Doc
from spacy.tokens.span import Span


class Vectorizer:

    @staticmethod
    def to_sents_df(doc: Doc, sent_window: int = 1, type_filter: Union[Set[str], Dict] = set(),
                    default_label: str = "NEG") -> pd.DataFrame:
        """
        Convert a SpaCy doc into pandas DataFrame. Assuming the doc has been labeled based on concepts(snippets), Vectorizer
        extends the input to the concepts' context sentences (depends on the sent_window size), generate labeled context
        sentences data, and return a pandas DataFrame (with three columns: 'X'---the text of context sentences,'concepts'---
        the text of labeled concepts, 'y'---label)
        @param doc: a SpaCy Doc
        @param sent_window: The window size (in sentences) around the target concept that need to be pulled
        @param type_filter: Specify whether and what types of annotation will be used generate the output DataFrame, this
        parameter can be defined as a set (only concept names are included) or a dictionary (where attributes and values
        can be included), which maps a matched concept (string and its context string) to a new value in "y"
        column in the output. The structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @param default_label: If there is no labeled concept in the context sentences, label it with this default_label
        @return: a pandas DataFrame
        """
        data_dict = Vectorizer.to_data_dict(doc, sent_window=sent_window, type_filter=type_filter,
                                            default_label=default_label, data_dict={'X': [], 'concept': [], 'y': []})
        df = pd.DataFrame(data_dict)
        return df

    @staticmethod
    def to_sents_nparray(doc: Doc, sent_window: int = 1, type_filter: Union[Set[str], Dict] = set(),
                         default_label: str = "NEG") -> ndarray:
        """
        Convert a SpaCy doc into numpy array. Assuming the doc has been labeled based on concepts(snippets), Vectorizer
        extends the input to the concepts' context sentences (depends on the sent_window size), generate labeled context
        sentences data, and return a numpy array (with three columns: 'X'---the text of context sentences,'concepts'---
        the text of labeled concepts, 'y'---label)
        @param doc: a SpaCy Doc
        @param sent_window: The window size (in sentences) around the target concept that need to be pulled
        @param type_filter: Specify whether and what types of annotation will be used generate the output array, this
        parameter can be defined as a set (only concept names are included) or a dictionary (where attributes and values
        can be included), which maps a matched concept (string and its context string) to a new value in "y"
        column in the output. The structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @param default_label: If there is no labeled concept in the context sentences, label it with this default_label
        @return: a numpy array
        """
        data_dict = Vectorizer.to_data_dict(doc, sent_window=sent_window, type_filter=type_filter,
                                            default_label=default_label, data_dict={'X': [], 'concept': [], 'y': []})
        rows = []
        for i in range(0, len(data_dict['X'])):
            rows.append([data_dict['X'][i], data_dict['concept'][i], data_dict['y'][i]])
        sents_nparray = np.array(rows)
        return sents_nparray

    @staticmethod
    def to_data_dict(doc: Doc, sent_window: int = 1, type_filter: Union[Set[str], Dict] = set(),
                     default_label: str = "NEG", data_dict: dict = {'X': [], 'concept': [], 'y': []}) -> Dict:
        """
        Convert a SpaCy doc into a labeled data dictionary. Assuming the doc has been labeled based on concepts(snippets), Vectorizer
        extends the input to the concepts' context sentences (depends on the sent_window size), generate labeled context
        sentences data, and return a dictionary (with three keys: 'X'---the text of context sentences,'concepts'---
        the text of labeled concepts, 'y'---label)
        @param doc: a SpaCy Doc
        @param sent_window: The window size (in sentences) around the target concept that need to be pulled
        @param type_filter: Specify whether and what types of annotation will be used generate the output DataFrame, this
        parameter can be defined as a set (only concept names are included) or a dictionary (where attributes and values
        can be included), which maps a matched concept (string and its context string) to a new value in "y"
        column in the output. The structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @param default_label: If there is no labeled concept in the context sentences, label it with this default_label
        @param data_dict: a dictionary to hold the output and pass on across documents, so that a corpus can be aggregated
        @param sent_idx: an IntervalTree built with all sentences in the doc
        @param context_sents: a 2-d list of sentences with predefined window size.
        @return: a dictionary
        """
        sent_idx = IntervalTree()
        sents = list(doc.sents)
        context_sents = []
        for i in range(0, len(sents) - sent_window + 1):
            begin_sent = sents[i]
            end_sent = sents[i + sent_window - 1]
            sent_idx.add(begin_sent.start, end_sent.end, len(context_sents))
            context_sents.append(sents[i:i + sent_window])

        concepts = []
        if hasattr(doc._, "concepts"):
            for type in doc._.concepts:
                if len(type_filter) == 0 or type in type_filter:
                    concepts.extend(doc._.concepts[type])
        else:
            concepts = [ent for ent in doc.ents if (len(type_filter) == 0 or ent.label in type_filter)]

        if isinstance(type_filter, Set):
            data_dict = Vectorizer.to_data_dict_on_types(concepts=concepts,
                                                         default_label=default_label,
                                                         data_dict=data_dict,
                                                         sent_idx=sent_idx, context_sents=context_sents)
        elif isinstance(type_filter, Dict):
            if len(type_filter) == 0:
                data_dict = Vectorizer.to_data_dict_on_types(concepts=concepts,
                                                             default_label=default_label,
                                                             data_dict=data_dict,
                                                             sent_idx=sent_idx, context_sents=context_sents)
            else:
                data_dict = Vectorizer.to_data_dict_on_type_attr_values(concepts=concepts,
                                                                        type_filter=type_filter,
                                                                        default_label=default_label,
                                                                        data_dict=data_dict,
                                                                        sent_idx=sent_idx, context_sents=context_sents)
        else:
            raise TypeError(
                'The arg: "type_filter" needs to be either a set of concept names or a dictionary. Not a {}:\n\t{}'.format(
                    type(type_filter), str(type_filter)))
        return data_dict

    @staticmethod
    def to_data_dict_on_types(concepts: List[Span],
                              default_label: str = "NEG", data_dict: dict = {'X': [], 'concept': [], 'y': []},
                              sent_idx: IntervalTree = None, context_sents: List[List[Span]] = None) -> Dict:
        """
        Convert a SpaCy doc into a labeled data dictionary. Assuming the doc has been labeled based on concepts(snippets), Vectorizer
        extends the input to the concepts' context sentences (depends on the sent_window size), generate labeled context
        sentences data, and return a dictionary (with three keys: 'X'---the text of context sentences,'concepts'---
        the text of labeled concepts, 'y'---label)
        @param concepts: a list of concepts (in Span type)
        @param default_label: If there is no labeled concept in the context sentences, label it with this default_label
        @param data_dict: a dictionary to hold the output and pass on across documents, so that a corpus can be aggregated
        @param sent_idx: an IntervalTree built with all sentences in the doc
        @param context_sents: a 2-d list of sentences with predefined window size.
        @return: a dictionary
        """
        if sent_idx is None or context_sents is None:
            return data_dict
        labeled_sents_id = set()
        for concept in concepts:
            context_sents_ids = sent_idx.search(concept.start, concept.end)
            for id in context_sents_ids:
                labeled_sents_id.add(id.data)
                context = context_sents[id.data]
                if concept.start >= context[0].start and concept.end <= context[-1].end:
                    data_dict['X'].append(' '.join([str(s) for s in context]))
                    data_dict['y'].append(concept.label_)
                    data_dict['concept'].append(str(concept))
        for i, context in enumerate(context_sents):
            if i not in labeled_sents_id:
                data_dict['X'].append(' '.join([str(s) for s in context]))
                data_dict['y'].append(default_label)
                data_dict['concept'].append('')
        return data_dict

    @staticmethod
    def to_data_dict_on_type_attr_values(concepts: List[Span], type_filter: Dict = dict(),
                                         default_label: str = "NEG",
                                         data_dict: dict = {'X': [], 'concept': [], 'y': []},
                                         sent_idx: IntervalTree = None, context_sents: List[List[Span]] = None) -> Dict:
        """
        Convert a SpaCy doc into a labeled data dictionary. Assuming the doc has been labeled based on concepts(snippets), Vectorizer
        extends the input to the concepts' context sentences (depends on the sent_window size), generate labeled context
        sentences data, and return a dictionary (with three keys: 'X'---the text of context sentences,'concepts'---
        the text of labeled concepts, 'y'---label)
        @param concepts: a list of concepts (in Span type)
        @param type_filter: Whether and what types of annotation with what attribute values will be used generate the
        output DataFrame, this parameter is defined as a dictionary (where attributes and values can be included), which
        maps a matched concept (string and its context string) to a new value in "y" column in the output. The
        structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @param default_label: If there is no labeled concept in the context sentences, label it with this default_label
        @param data_dict: a dictionary to hold the output and pass on across documents, so that a corpus can be aggregated
        @param sent_idx: an IntervalTree built with all sentences in the doc
        @param context_sents: a 2-d list of sentences with predefined window size.
        @return: a dictionary
        """
        if sent_idx is None or context_sents is None:
            return data_dict

        labeled_sents_id = set()
        for concept in concepts:
            conclusions = Vectorizer.get_mapped_names(concept=concept, type_filter=type_filter)
            if len(conclusions) > 0:
                context_sents_ids = sent_idx.search(concept.start, concept.end)
                for id in context_sents_ids:
                    labeled_sents_id.add(id.data)
                    context = context_sents[id.data]
                    if concept.start >= context[0].start and concept.end <= context[-1].end:
                        for conclusion in conclusions:
                            data_dict['X'].append(' '.join([str(s) for s in context]))
                            data_dict['y'].append(conclusion)
                            data_dict['concept'].append(str(concept))
        for i, context in enumerate(context_sents):
            if i not in labeled_sents_id:
                data_dict['X'].append(' '.join([str(s) for s in context]))
                data_dict['y'].append(default_label)
                data_dict['concept'].append('')
        return data_dict

    @staticmethod
    def get_mapped_names(concept: Span, type_filter: Dict) -> List[str]:
        """

        @param concept: a concept Span
        @param type_filter: a dictionary (where attributes and values can be included), which
        maps a matched concept (string and its context string) to a new value in "y" column in the output
        dictionary. The structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @return: a list of matched (on concept type, attribute and value pairs) new concept names
        """
        type_name = concept.label_
        value = type_filter[type_name]
        keys = []
        Vectorizer.get_mapped_name_by_attr_values(concept=concept, type_filter=value, keys=keys)
        return keys

    @staticmethod
    def get_mapped_name_by_attr_values(concept: Span, type_filter: Union[Dict, str] = None, keys: List[str] = []):
        """

        @param concept: a concept Span
        @param type_filter: a dictionary (where attributes and values can be included), which
        maps a matched concept (string and its context string) to a new value in "y" column in the output
        dictionary. The structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @param keys: a list of matched (on concept type, attribute and value pairs) new concept names that passed on
        during recursion
        """
        if type_filter is None or len(type_filter) == 0:
            return
        elif not isinstance(type_filter, Dict):
            keys.append(type_filter)
            return
        for attr in type_filter:
            if not hasattr(concept._, attr):
                return keys
            value = getattr(concept._, attr)
            if value in type_filter[attr]:
                Vectorizer.get_mapped_name_by_attr_values(concept=concept, type_filter=type_filter[attr][value],
                                                          keys=keys)
        pass

    @staticmethod
    def docs_to_sents_data_dict(docs: List[Doc], sent_window: int = 1, type_filter: Union[Set[str], Dict] = set(),
                                default_label: str = "NEG"):
        """
        Convert a list of SpaCy docs into a labeled data dictionary. Assuming the doc has been labeled based on concepts(snippets), Vectorizer
        extends the input to the concepts' context sentences (depends on the sent_window size), generate labeled context
        sentences data, and return a dictionary (with three keys: 'X'---the text of context sentences,'concepts'---
        the text of labeled concepts, 'y'---label)
        @param doc: a list of SpaCy Docs
        @param sent_window: The window size (in sentences) around the target concept that need to be pulled
        @param type_filter: Specify whether and what types of annotation will be used generate the output dictionary, this
        parameter can be defined as a set (only concept names are included) or a dictionary (where attributes and values
        can be included), which maps a matched concept (string and its context string) to a new value in "y"
        column in the output. The structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @param default_label: If there is no labeled concept in the context sentences, label it with this default_label
        @return: a dictionary
        """
        data_dict = {'X': [], 'concept': [], 'y': []}
        for doc in docs:
            Vectorizer.to_data_dict(doc, sent_window=sent_window, type_filter=type_filter,
                                    default_label=default_label, data_dict=data_dict)
        return data_dict

    @staticmethod
    def docs_to_sents_df(docs: List[Doc], sent_window: int = 1, type_filter: Union[Set[str], Dict] = set(),
                         default_label: str = "NEG") -> pd.DataFrame:
        """
        Convert a list of SpaCy docs into pandas DataFrame. Assuming the doc has been labeled based on concepts(snippets), Vectorizer
        extends the input to the concepts' context sentences (depends on the sent_window size), generate labeled context
        sentences data, and return a pandas DataFrame (with three columns: 'X'---the text of context sentences,'concepts'---
        the text of labeled concepts, 'y'---label)
        @param doc: a list of SpaCy Doc
        @param sent_window: The window size (in sentences) around the target concept that need to be pulled
        @param type_filter: Specify whether and what types of annotation will be used generate the output DataFrame, this
        parameter can be defined as a set (only concept names are included) or a dictionary (where attributes and values
        can be included), which maps a matched concept (string and its context string) to a new value in "y"
        column in the output. The structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @param default_label: If there is no labeled concept in the context sentences, label it with this default_label
        @return: a pandas DataFrame
        """
        data_dict = Vectorizer.docs_to_sents_data_dict(docs, sent_window=sent_window, type_filter=type_filter,
                                                       default_label=default_label)
        df = pd.DataFrame(data_dict)
        return df

    @staticmethod
    def docs_to_sents_nparray(docs: List[Doc], sent_window: int = 1, type_filter: Union[Set[str], Dict] = set(),
                              default_label: str = "NEG"):
        """
        Convert a list of SpaCy docs into numpy array. Assuming the doc has been labeled based on concepts(snippets), Vectorizer
        extends the input to the concepts' context sentences (depends on the sent_window size), generate labeled context
        sentences data, and return a numpy array (with three columns: 'X'---the text of context sentences,'concepts'---
        the text of labeled concepts, 'y'---label)
        @param doc: a list of SpaCy Doc
        @param sent_window: The window size (in sentences) around the target concept that need to be pulled
        @param type_filter: Specify whether and what types of annotation will be used generate the output array, this
        parameter can be defined as a set (only concept names are included) or a dictionary (where attributes and values
        can be included), which maps a matched concept (string and its context string) to a new value in "y"
        column in the output. The structure of expected dictionary will be:
        concept_type->attr1->value1->...(other attr->value pairs if needed)->mapped key name
        @param default_label: If there is no labeled concept in the context sentences, label it with this default_label
        @return: a numpy array
        """
        data_dict = Vectorizer.docs_to_sents_data_dict(docs, sent_window=sent_window, type_filter=type_filter,
                                                       default_label=default_label)
        rows = []
        for i in range(0, len(data_dict['X'])):
            rows.append([data_dict['X'][i], data_dict['concept'][i], data_dict['y'][i]])
        sents_nparray = np.array(rows)
        return sents_nparray



import logging
from collections import OrderedDict
from pathlib import Path
from typing import Union, Tuple, Type

from lxml import etree
from lxml.etree import Element, iterparse
from spacy.language import Language
from spacy.tokens.span import Span

from .base_reader import BaseDocReader, BaseDirReader


class EhostDocReader(BaseDocReader):
    """ This is a subclass of BaseDocReader to read in eHOST format files and generate SpaCy Docs
    """

    def __init__(self, nlp: Language = None, support_overlap: bool = False, use_adjudication: bool = False,
                 schema_file: Union[str, Path] = '', store_anno_string: bool = False,
                 log_level: bool = logging.WARNING):
        """

        :param nlp: a SpaCy language model
        :param support_overlap: if the EhostDocReader need to support reading from overlapped annotations.
            Because SpaCy's Doc.ents does not allows overlapped Spans, to support overlapping, Spans need to be stored
            somewhere else----Doc._.concepts
        :param use_adjudication: if read annotations from adjudication folder
        :param schema_file: initiate Span attributes using eHOST schema configuration file
        :param store_anno_string: whether read annotated string from annotations to double check parsed Span's correction
        :param log_level: set the logger's logging level. TO debug, set to logging.DEBUG
        """
        self.schema_set = False
        self.set_attributes(schema_file=schema_file)
        if store_anno_string:
            Span.set_extension("span_txt", default="")
        super().__init__(nlp=nlp, support_overlap=support_overlap, use_adjudication=use_adjudication,
                         store_anno_string=store_anno_string, log_level=log_level)
        pass

    def set_attributes(self, schema_file: Union[str, Path] = ''):
        """
        The current version SpaCy doesn't differentiate attributes for different annotation types.
        Thus, any attributes extended here will be applied to all Spans.
        :arg schema_file: initiate Span attributes using eHOST schema configuration file

        """
        schema_file = self.check_file_validity(schema_file, False)
        attr_names = set()
        if schema_file is not None:
            root = etree.parse(str(schema_file.absolute()))
            for attr_def in root.iter("attributeDef"):
                name = attr_def[0].text.replace(' ', '_')
                default_value = attr_def[2].text
                if name not in attr_names and not Span.has_extension(name):
                    Span.set_extension(name, default=default_value)
                    attr_names.add(name)
            self.schema_set = True
        pass

    def get_anno_content(self, txt_file: Path) -> str:
        """
        :arg txt_file from the text file path, infer knowtator.xml file path
        To use lxml's iterparse, just keep the path string here instead of reading the content
        """
        anno_file = self.infer_anno_file_path(txt_file)
        self.check_file_validity(anno_file)
        return str(anno_file.absolute())

    def infer_anno_file_path(self, txt_file: Path) -> Path:
        txt_file_name = txt_file.name
        anno_file_name = txt_file_name + '.knowtator.xml'
        if not self.use_adjudication:
            anno_file = Path(txt_file.parent.parent, 'saved', anno_file_name)
        else:
            anno_file = Path(txt_file.parent.parent, 'adjudication', anno_file_name)
        self.check_file_validity(anno_file)
        return anno_file

    def process_without_overlaps(self, doc):
        """
        Take in a SpaCy doc, add eHOST annotation(Span)s to doc.ents.
        This function doesn't support overlapped annotations.
         :arg doc: a SpaCy doc
        """
        sorted_span, classes, attributes = self.parse_to_dicts(self.anno, sort_spans=True)
        existing_entities = list(doc.ents)
        new_entities = list()
        token_left_bound = 0
        token_right_bound = len(doc) - 1
        token_start = -1
        token_end = -1
        for id, span_tuple in sorted_span:
            # because SpaCy uses token offset instead of char offset to define Spans, we need to match them,
            # binary search is used here to speed up
            if self.store_anno_string:
                start, end, span_txt = span_tuple
            else:
                start, end = span_tuple
            # because SpaCy uses token offset instead of char offset to define Spans, we need to match them,
            # binary search is used here to speed up
            if start < doc[0].idx:
                # If the annotation fall into a span that is before the 1st Spacy token, adjust the span to the 1st
                # token
                token_start = 0
                token_end = 1
            elif token_left_bound >= token_right_bound:
                # If the annotation fall into a span that is after the last Spacy token, adjust the span to the last
                # token
                token_start = token_right_bound - 2
                token_end = token_right_bound - 1
            else:
                token_start = self.find_start_token(start, token_left_bound, token_right_bound, doc)
                if end >= doc[-1].idx + doc[-1].__len__():
                    token_end = token_right_bound - 1
                else:
                    token_end = self.find_end_token(end, token_start, token_right_bound, doc)
            if token_start < 0 or token_start >= token_right_bound or token_end < 0 or token_end > token_right_bound:
                raise ValueError(
                    "It is likely your annotations overlapped, which process_without_overlaps doesn't support parsing "
                    "those. You will need to initiate the EhostDocReader with 'support_overlap=True' in the arguements")
            if token_start >= 0 and token_end > 0:
                span = Span(doc, token_start, token_end, label=classes[id][0])
                for attr_id in classes[id][1]:
                    attr_name = attributes[attr_id][0]
                    attr_value = attributes[attr_id][1]
                    setattr(span._, attr_name, attr_value)
                if self.store_anno_string and span_txt is not None:
                    setattr(span._, "span_txt", span_txt)
                new_entities.append(span)
                token_left_bound = token_end
            else:
                raise OverflowError(
                    'The span of the annotation: {}[{}:{}] is out of document boundary.'.format(classes[id][0], start,
                                                                                                end))
            pass
        doc.ents = existing_entities + new_entities
        return doc

    def process_support_overlaps(self, doc):
        """
        Take in a SpaCy doc, add eHOST annotation(Span)s. This function supports adding overlapped annotations.
         :arg doc: a SpaCy doc
        """
        sorted_span, classes, attributes = self.parse_to_dicts(self.anno, sort_spans=True)
        existing_concepts: dict = doc._.concepts
        token_left_bound = 0
        previous_abs_end = 0
        token_right_bound = len(doc) - 1
        token_start = -1
        token_end = -1
        for id, span_tuple in sorted_span:
            # because SpaCy uses token offset instead of char offset to define Spans, we need to match them,
            # binary search is used here to speed up
            if self.store_anno_string:
                start, end, span_txt = span_tuple
            else:
                start, end = span_tuple
            if start < doc[0].idx:
                # If the annotation fall into a span that is before the 1st Spacy token, adjust the span to the 1st
                # token
                token_start = 0
                token_end = 1
            elif token_left_bound >= token_right_bound:
                # If the annotation fall into a span that is after the last Spacy token, adjust the span to the last
                # token
                token_start = token_right_bound - 1
                token_end = token_right_bound
            else:
                if start < previous_abs_end:
                    token_start = self.find_start_token(start, token_start - 1 if token_start > 0 else 0,
                                                        token_right_bound, doc)
                    self.logger.debug("\tfind start token {}('{}')".format(token_start, doc[token_start]))
                else:
                    token_start = self.find_start_token(start, token_left_bound, token_right_bound, doc)
                    self.logger.debug("\tfind start token {}('{}')".format(token_start, doc[token_start]))
                if end >= doc[-1].idx + doc[-1].__len__():
                    token_end = token_right_bound
                else:
                    token_end = self.find_end_token(end, token_start, token_right_bound, doc)
                    self.logger.debug("\tfind end token {}('{}')".format(token_end, doc[token_end]))
            if token_start >= 0 and token_end > 0:
                span = Span(doc, token_start, token_end, label=classes[id][0])
                for attr_id in classes[id][1]:
                    attr_name = attributes[attr_id][0]
                    attr_value = attributes[attr_id][1]
                    setattr(span._, attr_name, attr_value)
                if self.store_anno_string and span_txt is not None:
                    setattr(span._, "span_txt", span_txt)
                if classes[id][0] not in existing_concepts:
                    existing_concepts[classes[id][0]] = list()
                existing_concepts[classes[id][0]].append(span)
                token_left_bound = token_end
                previous_abs_end = end

            else:
                raise OverflowError(
                    'The span of the annotation: {}[{}:{}] is out of document boundary.'.format(classes[id][0], start,
                                                                                                end))
            pass
        return doc

    def parse_to_dicts(self, xml_file: str, sort_spans: bool = False) -> Tuple[OrderedDict, OrderedDict]:
        iter = etree.iterparse(xml_file, events=('start',))
        # this doesn't seem elegant, but is said to be the fastest way
        spans = OrderedDict()
        classes = dict()
        attributes = OrderedDict()
        for event, ele in iter:
            if ele.tag == 'annotation':
                id, start, end, span_text = self.parse_annotation_tag(ele, iter)
                if self.store_anno_string:
                    spans[id] = (start, end, span_text)
                else:
                    spans[id] = (start, end)
            elif ele.tag == 'stringSlotMention':
                attr_id, attr, value = self.parse_attribute_tag(ele, iter)
                attributes[attr_id] = (attr, value)
            elif ele.tag == 'classMention':
                # <classMention id="EHOST_Instance_1">
                #   <hasSlotMention id="EHOST_Instance_8"/>
                #   <mentionClass id="Purulent">pain</mentionClass>
                # some annotations don't have "hasSlotMention" element
                # </classMention>
                id = ele.get('id')
            elif ele.tag == 'mentionClass':
                class_tag = ele.get('id')
                if id not in classes:
                    classes[id] = ['', []]
                classes[id][0] = class_tag
            elif ele.tag == 'hasSlotMention':
                if id not in classes:
                    classes[id] = ['', []]
                classes[id][1].append(ele.get('id'))
        if sort_spans:
            spans = sorted(spans.items(), key=lambda x: x[1][0])
        return spans, classes, attributes

    # <annotation>
    #   <mention id="EHOST_Instance_1"/>
    #   <annotator id="eHOST_2010">sjl</annotator>
    #   <span start="232" end="236"/>
    #   <spannedText>pain</spannedText>
    #   <creationDate>Sun Apr 19 19:30:11 MDT 2020</creationDate>
    # </annotation>
    def parse_annotation_tag(self, ele: Element, iter: iterparse) -> Tuple:
        id = None
        start = -1
        end = -1
        span_text = None
        for i in range(0, 4):
            eve, child = iter.__next__()
            if child.tag == 'mention':
                id = child.get('id')
            elif child.tag == 'span':
                start = int(child.get('start'))
                end = int(child.get('end'))
            elif child.tag == 'spannedText':
                span_text = child.text
        return id, start, end, span_text

    # <stringSlotMention id="EHOST_Instance_8">
    #   <mentionSlot id="status"/>
    #   <stringSlotMentionValue value="present"/>
    # </stringSlotMention>
    def parse_attribute_tag(self, ele: Element, iter: iterparse) -> Tuple:
        id = ele.get('id')
        attr = ''
        value = ''
        for i in range(0, 2):
            eve, child = iter.__next__()
            if child.tag == 'mentionSlot':
                attr = child.get('id')
            else:
                value = child.get('value')
        return id, attr, value


class EhostDirReader(BaseDirReader):
    def __init__(self, txt_dir: Union[str, Path], txt_extension: str = 'txt',
                 nlp: Language = None, docReaderClass: Type = None, support_overlap: bool = False,
                 recursive: bool = False, use_adjudication: bool = False,
                 schema_file: Union[str, Path] = '', **kwargs):
        """
        :param txt_dir: the directory contains text files (can be annotation file, if the text content and annotation
        content are saved in the same file). :param txt_extension: the text file extension name (default is 'txt').
        :param nlp: a SpaCy language model :param docReaderClass: a DocReader class that can be initiated. :param
        recursive: whether read file recursively down to the subdirectories. :param use_adjudication: if read
        annotations from adjudication folder :param schema_file: initiate Span attributes using eHOST schema
        configuration file
        """
        super().__init__(txt_dir=txt_dir, support_overlap=support_overlap, txt_extension=txt_extension, nlp=nlp,
                         docReaderClass=docReaderClass, recursive=recursive, use_adjudication=use_adjudication,
                         schema_file=schema_file, **kwargs)
        pass
